pub mod dm_type;
pub mod dvm_value;
pub mod lib;
pub mod mem;
pub mod object;
pub mod object_type;
pub mod op_codes;
pub mod platform;
pub mod util;

use crate::vm::dvm_value::DvmValue;
use crate::vm::lib::{DmConstant, DmLib};
use crate::vm::mem::{read_field_by_index, write_field_by_index};
use crate::vm::object::DvmObject;
use crate::vm::object_type::DmFn;
use op_codes::*;
use std::alloc::{alloc, dealloc, Layout};
use std::collections::HashMap;
use std::rc::Rc;

pub type PlatformFunction =
    fn(args: Vec<DvmValue>, state: &mut DvmState, context: &DvmContext) -> DvmValue;

#[derive(Debug)]
enum CallFrame {
    PlatformCall(CallFrameInfo),
    DeimosCall(CallFrameInfo),
}

#[derive(Debug)]
struct CallFrameInfo {
    pub fqn: String,
}

pub struct DvmContext {
    libs: Vec<Rc<DmLib>>,
    function_cache: HashMap<String, Rc<DmFn>>,
    platform_functions: HashMap<String, PlatformFunction>,
}

impl DvmContext {
    pub fn new() -> Self {
        DvmContext {
            libs: Vec::new(),
            function_cache: HashMap::new(),
            platform_functions: HashMap::new(),
        }
    }

    pub fn load_libs(&mut self, libs: Vec<Rc<DmLib>>) {
        for lib in libs {
            for lib_fn in lib.functions() {
                self.function_cache
                    .insert(lib_fn.fqn().to_string(), lib_fn.clone());
            }
            for interface_function in lib
                .interfaces()
                .iter()
                .flat_map(|interface| interface.functions())
            {
                self.function_cache.insert(
                    interface_function.fqn().to_string(),
                    interface_function.clone(),
                );
            }
            for implementation_function in lib
                .implementations()
                .iter()
                .flat_map(|implementation| implementation.functions())
            {
                self.function_cache.insert(
                    implementation_function.fqn().to_string(),
                    implementation_function.clone(),
                );
            }
            for method in lib
                .implementations()
                .iter()
                .flat_map(|implementation| implementation.methods())
            {
                self.function_cache
                    .insert(method.dm_fn().fqn().to_string(), method.dm_fn());
            }
            self.libs.push(lib);
        }
    }

    pub fn load_platform_fns(&mut self, platform_fns: HashMap<String, PlatformFunction>) {
        platform_fns.iter().for_each(|(fqn, platform_function)| {
            self.platform_functions
                .insert(fqn.to_string(), *platform_function);
        });
    }

    pub fn fn_by_fqn(&self, fqn: &str) -> Option<Rc<DmFn>> {
        self.function_cache.get(fqn).cloned()
    }
}

pub struct DvmState {
    call_stack: Vec<CallFrame>,
    registers: Vec<DvmValue>,
    register_state_stack: Vec<Vec<DvmValue>>,
}

impl DvmState {
    pub fn new() -> Self {
        DvmState {
            call_stack: Vec::new(),
            registers: Vec::new(),
            register_state_stack: Vec::new(),
        }
    }
}

macro_rules! dump_state {
    ( $state: expr ) => {
        println!("Registers: ");
        for (i, register) in $state.registers.iter().enumerate() {
            println!("  r{}: {:?}", i, register);
        }

        println!("Call stack:");
        for call_frame in &$state.call_stack {
            match call_frame {
                CallFrame::PlatformCall(frame_info) => {
                    println!("  {}", frame_info.fqn);
                }
                CallFrame::DeimosCall(frame_info) => {
                    println!("  {}", frame_info.fqn);
                }
            }
        }
    };
}

macro_rules! dvm_panic {
    ( $message: expr, $state: expr ) => {{
        println!("----");
        println!("Dvm panic: {}", $message);
        println!("----");
        dump_state!($state);
        panic!();
    }};
}

pub fn call_fn(
    state: &mut DvmState,
    context: &DvmContext,
    dm_fn: &DmFn,
    args: Vec<DvmValue>,
) -> DvmValue {
    // save current state
    state.call_stack.push(CallFrame::DeimosCall(CallFrameInfo {
        fqn: dm_fn.fqn().to_string(),
    }));
    state.register_state_stack.push(state.registers.clone());

    // zero registers and make sure there are enough for dm_function
    state.registers.clear();
    if state.registers.len() < dm_fn.number_used_registers() {
        state
            .registers
            .resize(dm_fn.number_used_registers(), DvmValue::Uninit);
    }

    // push args
    for (i, arg) in args.iter().enumerate() {
        state.registers[i] = arg.clone();
    }

    // run the byte code
    run_byte_code(state, context, &dm_fn.byte_code());

    // restore state
    state.registers = state.register_state_stack.pop().unwrap();
    state.call_stack.pop();

    // get return value
    if let Some(return_register) = dm_fn.return_register() {
        state
            .registers
            .get(return_register as usize)
            .map(|m| m.clone())
            .unwrap()
    } else {
        DvmValue::Void
    }
}

macro_rules! next_8 {
    ( $byte_code: expr, $T: ident ) => {
        $byte_code.next().unwrap() as $T
    };
}

macro_rules! next_32_le {
    ( $byte_code: expr, $T: ident ) => {
        next_8!($byte_code, $T)
            + (next_8!($byte_code, $T) << 8)
            + (next_8!($byte_code, $T) << 16)
            + (next_8!($byte_code, $T) << 24)
    };
}

macro_rules! next_64_le {
    ( $byte_code: expr, $T: ident ) => {
        next_8!($byte_code, $T)
            + (next_8!($byte_code, $T) << 8)
            + (next_8!($byte_code, $T) << 16)
            + (next_8!($byte_code, $T) << 24)
            + (next_8!($byte_code, $T) << 32)
            + (next_8!($byte_code, $T) << 40)
            + (next_8!($byte_code, $T) << 48)
            + (next_8!($byte_code, $T) << 56)
    };
}

macro_rules! next_usize_le {
    ( $byte_code: expr ) => {{
        let mut u: usize = 0;
        for i in 0..std::mem::size_of::<usize>() {
            u += next_8!($byte_code, usize) << (8 * i);
        }
        u
    }};
}

macro_rules! read_array_u8 {
    ( $iter: expr ) => {{
        let arr_length = next_8!($iter, u8);
        let mut arr = Vec::with_capacity(arr_length as usize);
        let mut i = 0;
        while i < arr_length {
            arr.push(next_8!($iter, u8));
            i += 1;
        }
        arr
    }};
}

macro_rules! read_string {
    ( $iter: expr ) => {{
        let str_length = next_32_le!($iter, usize);
        let mut str_raw: Vec<u8> = Vec::new();
        let mut i = 0;
        while i < str_length {
            str_raw.push($iter.next().unwrap());
            i += 1;
        }
        String::from_utf8(str_raw).unwrap()
    }};
}

pub fn run_byte_code(state: &mut DvmState, context: &DvmContext, byte_code: &[u8]) {
    let mut iter = byte_code.iter().cloned();
    while let Some(op_code) = iter.next() {
        match op_code {
            MOV_BYTE => {
                todo!()
            }
            MOV_INT => {
                let target_register = next_8!(iter, usize);
                let operand = next_32_le!(iter, u32) as i32;
                state.registers[target_register] = DvmValue::Int(operand);
            }
            MOV_LONG => {
                let target_register = next_8!(iter, usize);
                let operand = next_64_le!(iter, u64) as i64;
                state.registers[target_register] = DvmValue::Long(operand);
            }
            MOV_DOUBLE => {
                let target_register = next_8!(iter, usize);
                let operand = next_64_le!(iter, u64) as f64;
                state.registers[target_register] = DvmValue::Double(operand);
            }
            MOV_REGISTER => {
                let target_register = next_8!(iter, usize);
                let source_register = next_8!(iter, usize);
                let source_value =
                    std::mem::replace(&mut state.registers[source_register], DvmValue::Uninit);
                state.registers[target_register] = source_value;
            }
            LOAD => {
                let target_register = next_8!(iter, usize);
                let source_register = next_8!(iter, usize);
                let source_object_field_index = next_usize_le!(iter);

                let source_object = state
                    .registers
                    .get(source_register)
                    .unwrap()
                    .expect_object();

                state.registers[target_register] =
                    read_field_by_index(source_object_field_index, &source_object);
            }
            STORE => {
                let target_register = next_8!(iter, usize);
                let target_object_field_index = next_usize_le!(iter);
                let source_register = next_8!(iter, usize);

                let target_object = state
                    .registers
                    .get(target_register)
                    .unwrap()
                    .expect_object();

                let source_value =
                    std::mem::replace(&mut state.registers[source_register], DvmValue::Uninit);
                write_field_by_index(target_object_field_index, &target_object, source_value);
            }
            ALLOC => {
                let target_register = next_8!(iter, usize);
                let impl_name = read_string!(iter);

                let implementation = context
                    .libs
                    .iter()
                    .find_map(|lib| {
                        lib.implementations()
                            .iter()
                            .find(|implementation| implementation.fqn() == impl_name)
                    })
                    .unwrap_or_else(|| {
                        dvm_panic!(&format!("Implementation not found: {}", impl_name), state)
                    })
                    .clone();

                let dm_alloc_object = DvmObject::new(implementation);
                state.registers[target_register] = DvmValue::from_object(dm_alloc_object);
            }
            DEALLOC => {
                let target_register = next_8!(iter, usize);
                let target_object =
                    std::mem::replace(&mut state.registers[target_register], DvmValue::Uninit)
                        .expect_object();
                drop(target_object); // explicit
                state.registers[target_register] = DvmValue::Uninit;
            }
            MOV_CONST => {
                // Decode
                let address_register = next_8!(iter, usize);
                let size_register = next_8!(iter, usize);
                let lib_name = read_string!(iter);
                let const_id = next_32_le!(iter, usize);

                // Get constant
                let Some(lib) = context.libs.iter().find(|lib| lib.name() == lib_name) else {
                    dvm_panic!(
                        &format!("Could not find lib with name: {}", lib_name),
                        state
                    );
                };
                let constant = lib.constants().get(const_id).unwrap();
                let DmConstant::String(s) = constant;
                let pointer = s.as_ptr();

                // Store the pointer metadata in target registers.
                state.registers[address_register] = DvmValue::USize(pointer as usize);
                state.registers[size_register] = DvmValue::USize(s.len());
            }
            ALLOC_RAW => {
                // Allocates a raw number of bytes, with the number of bytes determined by the value
                // of the source register.
                let address_register = next_8!(iter, usize);
                let source_register = next_8!(iter, usize);
                let size = state.registers.get(source_register).unwrap().expect_long() as usize;
                let layout = Layout::from_size_align(size, 1).unwrap();
                let raw_ptr = unsafe { alloc(layout) };
                state.registers[address_register] = DvmValue::USize(raw_ptr as usize);
            }
            DEALLOC_RAW => {
                // Deallocates a DvmValue::Pointer(DvmPointer) in the target register.
                let address_register = next_8!(iter, usize);
                let size_register = next_8!(iter, usize);
                let address = state
                    .registers
                    .get(address_register)
                    .unwrap()
                    .expect_usize();
                let size = state.registers.get(size_register).unwrap().expect_usize();
                let layout = Layout::from_size_align(size, 1).unwrap();
                unsafe {
                    dealloc(address as *mut u8, layout);
                }
                state.registers[address_register] = DvmValue::Uninit;
                state.registers[size_register] = DvmValue::Uninit;
            }
            PLATFORM_CALL => {
                // Calls a platform function. The result of the platform call is stored in the
                // specified return register.
                let symbol_name = read_string!(iter);
                let return_register = next_8!(iter, usize);
                let arg_registers = read_array_u8!(iter);

                let mut args = Vec::with_capacity(arg_registers.len());
                for arg_register in arg_registers {
                    let value = state.registers.get(arg_register as usize).unwrap();
                    args.push(value.clone());
                }

                let platform_function = context.platform_functions.get(&symbol_name).unwrap();
                state
                    .call_stack
                    .push(CallFrame::PlatformCall(CallFrameInfo { fqn: symbol_name }));
                let call_result = platform_function(args, state, context);
                state.registers[return_register] = call_result;
                state.call_stack.pop();
            }
            INVOKE_FN => {
                let symbol_name = read_string!(iter);
                let return_register = next_8!(iter, usize);
                let arg_registers = read_array_u8!(iter);

                let mut args = Vec::with_capacity(arg_registers.len());
                for arg_register in arg_registers {
                    let value = state.registers.get(arg_register as usize).unwrap();
                    args.push(value.clone());
                }

                let dm_fn = context.function_cache.get(&symbol_name).unwrap();
                let call_result = call_fn(state, context, dm_fn, args);
                state.registers[return_register] = call_result;
            }
            INVOKE_VIRTUAL => {
                let symbol_name = read_string!(iter);
                let return_register = next_8!(iter, usize);
                let arg_registers = read_array_u8!(iter);

                let mut args = Vec::with_capacity(arg_registers.len());
                for arg_register in arg_registers {
                    let value = state.registers.get(arg_register as usize).unwrap();
                    args.push(value.clone());
                }

                let self_obj = args.get(0).unwrap().expect_object();
                let method = self_obj
                    .implementation()
                    .find_method(&symbol_name, &self_obj)
                    .unwrap_or_else(|| {
                        dvm_panic!(
                            &format!("Could not find method with name: {}", symbol_name),
                            state
                        );
                    });
                let dm_fn = method.dm_fn();

                let call_result = call_fn(state, context, &dm_fn, args);
                state.registers[return_register] = call_result;
            }
            INVOKE_DYNAMIC => {
                unimplemented!("INVOKE_DYNAMIC is not yet supported and may never be.")
            }
            MULTIPLY => {
                todo!()
            }
            op_code => {
                dvm_panic!(
                    &format!("Unimplemented or unknown op code: {}", op_code),
                    state
                );
            }
        }
    }
}

#[cfg(test)]
mod run_code_tests {
    use super::*;
    use crate::vm::dm_type::DmType;
    use crate::vm::object_type::{DmField, DmImplementation};

    macro_rules! assert_register {
        ( $expected: expr, $state: expr, $register_number: expr ) => {
            assert_eq!($expected, $state.registers.remove($register_number));
        };
    }

    fn setup<'a>(number_of_registers: usize) -> (DvmState, DvmContext) {
        let mut state = DvmState::new();
        state
            .registers
            .resize(number_of_registers, DvmValue::Uninit);
        (state, DvmContext::new())
    }

    fn impl_with_object_field() -> DmImplementation {
        let mut dm_impl = DmImplementation::new("ImplWithObjectField", "ImplWithObjectField", None);
        let object_field = DmField::new("object_field", DmType::Object);
        dm_impl.add_field(object_field);
        dm_impl
    }

    #[test]
    fn mov_1_as_int() {
        let mut code = Vec::new();
        add_mov_int(&mut code, 0, 1);
        let (mut state, context) = setup(1);
        run_byte_code(&mut state, &context, &code);
        assert_register!(DvmValue::Int(1), state, 0);
    }

    #[test]
    fn move_65535_as_int() {
        let mut code = Vec::new();
        add_mov_int(&mut code, 0, 0xffff);
        let (mut state, context) = setup(1);
        run_byte_code(&mut state, &context, &code);
        assert_register!(DvmValue::Int(0xffff), state, 0);
    }

    #[test]
    fn move_int_max_as_int() {
        let mut code = Vec::new();
        add_mov_int(&mut code, 0, 0x0fff_ffff);
        let (mut state, context) = setup(1);
        run_byte_code(&mut state, &context, &code);
        assert_register!(DvmValue::Int(0x0fff_ffff), state, 0);
    }

    #[test]
    fn move_register() {
        let mut code = Vec::new();
        add_mov_int(&mut code, 1, 1);
        add_mov_register(&mut code, 0, 1);
        let (mut state, context) = setup(2);
        run_byte_code(&mut state, &context, &code);
        assert_register!(DvmValue::Int(1), state, 0);
    }

    #[test]
    fn load_object() {
        let dummy_impl = impl_with_object_field();
        let mut dummy_lib = DmLib::new("dummy");
        dummy_lib.add_implementation(dummy_impl);
        let dummy_impl_ref = dummy_lib
            .implementations()
            .iter()
            .find(|implementation| implementation.fqn() == "dummy")
            .unwrap();

        let mut code = Vec::new();
        add_alloc(&mut code, 0, dummy_impl_ref.fqn());
        add_alloc(&mut code, 1, dummy_impl_ref.fqn());
        add_store(&mut code, 0, 0, 1);
        add_load(&mut code, 2, 0, 0);

        let (mut state, mut context) = setup(3);
        state.registers.resize(3, DvmValue::Uninit);
        context.load_libs(vec![Rc::new(dummy_lib)]);
        run_byte_code(&mut state, &context, &code);

        assert!(state.registers[2].is_object())
    }
}
